@@ -109,6 +109,12 @@ tmpl_read(char **tmp, char *tmpl_file)
 
     if (*tmp == NULL) {
         log(ERR_TEMPLATE_READ, "tmpl_read", tmpl_file);
+#ifdef __CUSTOMIZE2018
+        // mark empty information
+    if (ismode_enc) {        
+        *tmp = "";
+    }
+#endif	// __CUSTOMIZE2018
         return(ERROR);
     }
 
