@@ -50,6 +50,13 @@
 #include "global.h"
 #include "mailzip_tmpl.h"
 #include "samma_policy.h"
+#ifdef __CUSTOMIZE2018
+#include <libmilter/mfapi.h>
+#include <syslog.h>
+#include <libdgmail.h>
+#include "mailzip.h"
+#define MLFIPRIV(ctx)        ((struct mlfiPriv *) smfi_getpriv(ctx))
+#endif	// __CUSTOMIZE2018
 
 pthread_mutex_t gmime_lock = PTHREAD_MUTEX_INITIALIZER;
 
@@ -128,6 +135,13 @@ mailzip_clean(struct mailzip mz)
     if (mz.attachfilename != NULL) {
         free(mz.attachfilename);
     }
+
+#ifdef __CUSTOMIZE2018
+    if (ismode_enc) {
+        // release addmsg_tmpl
+        addmsg_release(mz.am_tmpl);
+    }
+#endif	// __CUSTOMIZE2018
 }
 /*
  * delete_attachmetns_mail
@@ -258,8 +272,22 @@ zip_convert_mail(SMFICTX *ctx, struct mailinfo *minfo, struct config *cfg,
     struct rcptaddr *rdm_noext = NULL;
     struct rcptaddr *rdm_all = NULL;
     struct rcptinfo tmp_rcptinfo;
+#ifdef __CUSTOMIZE2018
+    // for password log
+    struct mlfiPriv *priv;
+    char *msgid;
+#endif	// __CUSTOMIZE2018
+#ifdef __NOTICE_PASSWD
+    char *notify_message_to;
+#endif // __NOTICE_PASSWD
 
     memset(&mz, 0, sizeof(struct mailzip));
+#ifdef __CUSTOMIZE2018
+    // initialize addmsg_tmpl
+    mz.am_tmpl = addmsg_init();
+    mz.am_priv = MLFIPRIV(ctx);
+#endif	// __CUSTOMIZE2018
+
 
     /* parse mail */
     ret = parse_mail(minfo, cfg, &mz);
@@ -296,6 +324,17 @@ zip_convert_mail(SMFICTX *ctx, struct mailinfo *minfo, struct config *cfg,
 		mailzip_clean(mz);
 		return ZIP_CONVERT_FAILED;
 	    }
+#ifdef __CUSTOMIZE2018
+            if (ismode_enc) {
+                // for password log
+                if ((priv = MLFIPRIV(ctx)) == NULL) {
+                    msgid = NULL;
+                } else {
+                    msgid = priv->mlfi_sendercheck_arg.message_id;
+                }
+                zipconv_log(msgid, from, p->rcptlist, mz.namelist, p->passwd);
+            }
+#endif	// __CUSTOMIZE2018
             /* send confirmation mail */
             if ((cfg->cf_fixedpassnotify[0] == 'Y' || cfg->cf_fixedpassnotify[0] == 'y')) {
                 org_p = p->passwd;	/* save original password */
@@ -390,6 +429,17 @@ zip_convert_mail(SMFICTX *ctx, struct mailinfo *minfo, struct config *cfg,
                     free_rcptlist(rdm_all);
 	            return ZIP_CONVERT_FAILED;
 	        }
+#ifdef __CUSTOMIZE2018
+                if (ismode_enc) {
+                    // for password log
+                    if ((priv = MLFIPRIV(ctx)) == NULL) {
+                        msgid = NULL;
+                    } else {
+                        msgid = priv->mlfi_sendercheck_arg.message_id;
+                    }
+                    zipconv_log(msgid, from, p->rcptlist, mz.namelist, p->passwd);
+                }
+#endif	// __CUSTOMIZE2018
             } else {
                 /* 拡張子が指定されていない場合 */
                 /* まとめて送信するためのリストを作成 */
@@ -446,6 +496,17 @@ zip_convert_mail(SMFICTX *ctx, struct mailinfo *minfo, struct config *cfg,
                 free_rcptlist(rdm_all);
                 return ZIP_CONVERT_FAILED;
             }
+#ifdef __CUSTOMIZE2018
+            if (ismode_enc) {
+                // for password log
+                if ((priv = MLFIPRIV(ctx)) == NULL) {
+                    msgid = NULL;
+                } else {
+                    msgid = priv->mlfi_sendercheck_arg.message_id;
+                }
+                zipconv_log(msgid, from, tmp_rcptinfo.rcptlist, mz.namelist, tmp_rcptinfo.passwd);
+            }
+#endif	// __CUSTOMIZE2018
             free_rcptlist(rdm_noext);
         }
 
@@ -492,12 +553,25 @@ zip_convert_mail(SMFICTX *ctx, struct mailinfo *minfo, struct config *cfg,
             }
         }
 
-	if (passwd_sendmail(cfg, from, notify_message) != 0) {
-	    free(notify_message);
-	    mailzip_clean(mz);
+        /* if passwordnotice item set is 0 or 2 
+         * then send passwordnotice mail to FROM */
+#ifdef __NOTICE_PASSWD
+        if ((cfg->cf_passwordnotice == 0) || (cfg->cf_passwordnotice == 2)) {
+	    if (passwd_sendmail(cfg, from, notify_message) != 0) {
+	        free(notify_message);
+	        mailzip_clean(mz);
+                free_rcptlist(rdm_all);
+	        return ZIP_CONVERT_FAILED;
+	    }
+        }
+#else  // __NOTICE_PASSWD
+        if (passwd_sendmail(cfg, from, notify_message) != 0) {
+            free(notify_message);
+            mailzip_clean(mz);
             free_rcptlist(rdm_all);
-	    return ZIP_CONVERT_FAILED;
-	}
+            return ZIP_CONVERT_FAILED;
+        }
+#endif
 	free(notify_message);
 
         for (i = 0; (rdm_all + i)->rcpt_addr; i++) {
@@ -511,6 +585,37 @@ zip_convert_mail(SMFICTX *ctx, struct mailinfo *minfo, struct config *cfg,
             }
         }
         free_rcptlist(rdm_all);
+
+#ifdef __NOTICE_PASSWD
+        for (p = rdmpasslist; p != NULL; p = p->Next) {
+            /* loop all to then send passwordnotice */
+            for (i = 0; (p->rcptlist + i)->rcpt_addr != NULL; i++) {
+                rcptaddr = (p->rcptlist + i)->rcpt_addr;
+
+                /* if passwordnotice item set is 1 or 2 then send passwordnotice mail to TO */
+                if ((cfg->cf_passwordnotice == 1) || (cfg->cf_passwordnotice == 2)) {
+                    /* make notify message */
+                    temp = tmpl_init();
+                    if (tmpl_tag_replace(temp, mz, rdmpasslist, rcptaddr, &notify_message_to) != 0){
+                        tmpl_release(temp);
+                        mailzip_clean(mz);
+                        return ZIP_CONVERT_FAILED;
+                    }
+                    tmpl_release(temp);
+
+                    /* send notify_password to TO */
+                    if (passwd_sendmail(cfg, rcptaddr, notify_message_to) != 0) {
+                        free(notify_message_to);
+                        mailzip_clean(mz);
+                        return ZIP_CONVERT_FAILED;
+                    }
+
+                    /* free notify_message_to */
+                    free(notify_message_to);
+                }
+            }
+        }
+#endif //__NOTICE_PASS
     } 
 
 #ifdef	DEBUG
@@ -1100,7 +1205,9 @@ drop_file(GMimeObject *object, struct config *cfg, struct mailzip *mz)
     char tmppath[PATH_MAX + 1];
     int count = 0;
     FILE *wfp;
-
+#ifdef __CUSTOMIZE2018
+    GMimeContentType *ctype;
+#endif	// __CUSTOMIZE2018
 
     /* get attach file number */
 #ifndef GMIME26
@@ -1108,14 +1215,26 @@ drop_file(GMimeObject *object, struct config *cfg, struct mailzip *mz)
 #else
     index = g_mime_multipart_get_count((GMimeMultipart *)object);
 #endif
-
     for (i = index - 1; i >= 0; i--) {
 	att = g_mime_multipart_get_part((GMimeMultipart *)object, i);
 
 	if (GMIME_IS_MESSAGE_PART(att)) {
 	    filename = MESSAGE_FILENAME;
 	} else if (GMIME_IS_MULTIPART (att)) {
+#ifdef __CUSTOMIZE2018
+            if (ismode_enc) {
+                // store position and depth
+                mz->md_depth ++;
+                mz->md_pos = i;
+            }
+#endif	// __CUSTOMIZE2018
 	    ret = drop_file(att, cfg, mz);
+#ifdef __CUSTOMIZE2018
+            if (ismode_enc) {
+                // restore depth
+                mz->md_depth --;
+            }
+#endif	// __CUSTOMIZE2018
 
 #ifndef GMIME24
 	    g_object_unref (att);
@@ -1131,7 +1250,30 @@ DEBUGLOG("getting filename...");
 
 	    /* get filename */
 	    filename = g_mime_part_get_filename((GMimePart *)att);
+#ifdef __CUSTOMIZE2018
+            if (ismode_enc) {
+                // Chcek if this part is to be added the message
+                if (*(cfg->cf_useaddmessageheader) == 'Y'
+                    || *(cfg->cf_useaddmessageheader) == 'y') {
+                    ctype = g_mime_object_get_content_type(att);
+                    if (ctype->type != NULL
+                            && strcasecmp(ctype->type, "text") == 0) {
+                        // case text part
+                        if ((mz->md_depth == 0 && i == 0)
+                             || (mz->md_depth == 1 && mz->md_pos == 0)) {
+                            // case first postion
+                            ret = replace_mime_part(att, object, i, mz, ctype);
+                            // ZZ-ERROR SHORI
+                            if (ret != 0) {
+                                return -1;
+                            }
+                        }
+                    }
 
+                }
+            }
+// g_mime_object_append_header (att, "hoge", "hage");
+#endif	// __CUSTOMIZE2018
 	    if (filename == NULL) {
 DEBUGLOG("Failed to get filename when index = %d", i);
 #ifndef GMIME24 
@@ -1661,3 +1803,505 @@ int set_references(char *mail_data, char **passwd_mail) {
 
     return SUCCESS;
 }
+
+#ifdef __CUSTOMIZE2018
+/* zipconv_log
+ *
+ * output log with message-id, sender, rcpts, attached files, and password
+ *
+ * Args:
+ *    char *message_id			Message-Id
+ *    char *from			Sener address
+ *    struct rcptaddr *rcpts		Recipients
+ *    struct name_list *attachlist	File name list
+ *    char *passwd			password
+ *
+ * Return:
+ *     (void)
+ */
+void
+zipconv_log(char *message_id, char *from, struct rcptaddr *rcpts,
+		struct name_list *attachlist, char *passwd)
+{
+    size_t loglen = 15;	// length : 16 = FMT + '\0'
+    char *msgid = NULL;
+    char *fromstr = NULL;
+    char *pwstr = NULL;
+    char *log_msg = NULL;
+    size_t msgid_len = 0;
+    size_t from_len = 0;
+    size_t passwd_len = 0;
+    int i;
+    int oft = 0;
+    char *new_log = NULL;
+    int ret;
+
+    // sanitize
+    msgid = (message_id == NULL)?"-":message_id;
+    fromstr = (from == NULL)?"-":from;
+    pwstr = (passwd == NULL)?"-":passwd;
+        
+    // calc log length
+    msgid_len = strlen(msgid);
+    loglen += msgid_len;
+    from_len = strlen(fromstr);
+    loglen += from_len;
+    for (i = 0; (rcpts + i)->rcpt_addr != NULL; i++) {
+        loglen += (rcpts + i)->rcpt_addr_len + 1;	// includes ':'
+    }
+    for (i = 0; (attachlist + i)->attach_name != NULL; i++) {
+        loglen += (attachlist + i)->attach_name_len + 1;	// includes ':'
+    }
+    passwd_len = strlen(pwstr);
+    loglen += passwd_len;
+
+    // allocate memory
+    log_msg = (char *)malloc(loglen);
+    if (log_msg == NULL) {
+        log(ERR_MEMORY_ALLOCATE, "zipconv_log", "log_msg", strerror(errno));
+        return;
+    }
+    log_msg[0] = '\0';
+
+    // build log message
+    // Note:
+    //  To avoid to reduce performance, we don't use sprintf or strcat, ...
+
+    // message-id
+    strcpy((log_msg + oft), "\"");
+    oft ++;
+    strcpy((log_msg + oft), msgid);
+    oft += msgid_len;
+    strcpy((log_msg + oft), "\" \"");
+    oft += 3;
+
+    // from addr
+    strcpy((log_msg + oft), fromstr);
+    oft += from_len;
+    strcpy((log_msg + oft), "\" \"");
+    oft += 3;
+
+    // to addr
+    for (i = 0; (rcpts + i)->rcpt_addr != NULL; i++) {
+        if (i != 0) {
+            // add delimiter
+            *(log_msg + oft) = ':';
+            oft ++;
+        }
+        strcpy((log_msg + oft), (rcpts + i)->rcpt_addr);
+        oft += (rcpts + i)->rcpt_addr_len;
+    }
+    strcpy((log_msg + oft), "\" \"");
+    oft += 3;
+
+    // filename
+    for (i = 0; (attachlist + i)->attach_name != NULL; i++) {
+        if (i != 0) {
+            // add delimiter
+            *(log_msg + oft) = ':';
+            oft ++;
+        }
+        strcpy((log_msg + oft), (attachlist + i)->attach_name);
+        oft += (attachlist + i)->attach_name_len;
+    }
+    strcpy((log_msg + oft), "\" \"");
+    oft += 3;
+
+    // message-id
+    strcpy((log_msg + oft), pwstr);
+    oft += passwd_len;
+    strcpy((log_msg + oft), "\"");
+    oft ++;
+
+    // change encoding
+    ret = dg_str2code(log_msg, &new_log, "SJIS", "UTF-8");
+    switch (ret) {
+    case 0:
+        free(log_msg);
+        log_msg = new_log;
+        break;
+    default:
+        // ignore silently if error occurs
+        if (new_log != NULL) {
+            free(new_log);
+        }
+        break;
+    }
+
+    // output log
+    if ((void *)log == errorlog) {
+        // output to stderr
+        strcpy((log_msg + oft), "\n");
+        fprintf(stderr, log_msg);
+    } else if ((void*)log == systemlog) {
+        syslog(LOG_INFO, log_msg);
+    }
+    free(log_msg);
+    return;
+}
+
+/* get_body_head
+ *
+ * search the position of the body head
+ *
+ * Args:
+ *    char *buf	String pointer to search the position of the body head
+ *
+ * Return:
+ *    char *	The body head pointer
+ */
+char *
+get_body_head(char *buf)
+{
+    char *head;
+    char *p1, *p2;
+    if (*buf == '\0') {
+        // case no string
+        head = buf;
+    } else if (*buf == '\n') {
+        // case no header with '\n'
+        head = buf + 1;
+    } else if (*buf == '\r' || *buf == '\n') {
+        // case no header with '\r\n'
+        head = buf + 2;
+    } else {
+        p1 = strstr(buf, "\n\n");
+        p2 = strstr(buf, "\r\n\r\n");
+        if (p1 == NULL) {
+            if (p2 == NULL) {
+                // case no blank line
+                head = buf;
+            } else {
+                // found blank line with '\r\n'
+                head = p2 + 4;
+            }
+        } else {
+            if (p2 == NULL) {
+                // found blank line with '\n'
+                head = p1 + 2;
+            } else {
+                // both line feed found
+                if (p1 > p2) {
+                    // blank line is with '\r\n'
+                    head = p2 + 4;
+                } else {
+                    // blank line is with '\n'
+                    head = p1 + 2;
+                }
+            }
+        }
+    }
+    return head;
+}
+
+/* is_noenc_qp
+ *
+ * evaluate the charactor necessary to encode QP
+ *
+ * Args:
+ *    char c	to evaluate charactor
+ *
+ * Return:
+ *    1		not necessary
+ *			printable without '='
+ *			'\r' '\n'
+ *    0		necessary
+ */
+int
+is_noenc_qp(char c)
+{
+    if ((c >= ' ' && c <= '<') || (c >= '>' && c <= '~')
+         || c == '\r' || c == '\n') {
+        return 1;
+    } else {
+        return 0;
+    }
+}
+
+/* encode_qp
+ *
+ * encode the string to quoted-printable
+ *
+ * Args:
+ *    char *src	to encode string
+ *
+ * Return:
+ *    char *	encoded string allocated
+ *    NULL	memory error
+ */
+char *
+encode_qp(char *src)
+{
+    size_t len;
+    int soft = 0, doft = 0;
+    char *new_buf;
+    char one_char[4];
+
+    len = strlen(src);
+    new_buf = (char *)malloc(len * 3 + 1);
+    if (new_buf == NULL) {
+        return NULL;
+    }
+    *new_buf = '\0';
+
+    for (soft = 0, doft = 0; src[soft] != '\0'; soft ++) {
+        //if (isascii(src[soft])) {
+        if (is_noenc_qp(src[soft])) {
+            *(new_buf + doft) = src[soft];
+            *(new_buf + doft + 1) = '\0';
+            doft ++;
+        } else {
+            snprintf(one_char, 4, "=%02x", ((unsigned int)src[soft] & 0xff));
+            strcpy(new_buf + doft, one_char);
+            doft += 3;
+        }
+    }
+    return new_buf;
+}
+
+/* add_message
+ *
+ * add the template message to the top of the original message
+ *
+ * Args:
+ *    char *orig_msg	original message
+ *    struct mailzip *mz	mailzip structure
+ *    char *charset	Charactor set
+ *    GMIMEContentEncoding	c_enc	encode code
+ *		GMIME_CONTENT_ENCODING_BASE64
+ *		GMIME_CONTENT_ENCODING_QUOTEDPRINTABLE
+ *		and other
+ *
+ * Return:
+ *    char *	merged string allocated
+ *    NULL	memory error
+ */
+char *
+add_message(char *orig_msg, struct mailzip *mz,
+            char *charset, GMimeContentEncoding c_enc, GMimeContentType *ctype)
+{
+    char *tmpl = NULL;
+    char *tmp_tmpl = NULL, *new_tmpl = NULL;
+    char *tmp_msg = NULL, *tmp_msg2 = NULL, *new_msg = NULL;
+    int ret;
+    size_t olen = 0, tlen = 0, nlen = 0;
+    int hflag;
+
+    hflag = (strcasecmp(ctype->subtype, "html") == 0)?1:0;
+
+    switch(mz->am_priv->mlfi_subject_lang) {
+    case SBJLANG_JP:
+        tmpl = (hflag == 0)?mz->am_tmpl->amt_jp:mz->am_tmpl->amt_jp_html;
+        break;
+    case SBJLANG_EN:
+        tmpl = (hflag == 0)?mz->am_tmpl->amt_en:mz->am_tmpl->amt_en_html;
+        break;
+    default:
+        tmpl = (hflag == 0)?mz->am_tmpl->amt_both:mz->am_tmpl->amt_both_html;
+        break;
+    }
+
+    if (tmpl == NULL || *tmpl == '\0') {
+        // case empty template
+        new_msg = strdup(orig_msg);
+        if (new_msg == NULL) {
+	    log(ERR_MEMORY_ALLOCATE, "add_message", "new_msg", strerror(errno));
+            return NULL;
+        }
+        return new_msg;
+    }
+
+    // encode template
+    if (charset == NULL) {
+        // case no charset force EN
+        tmpl = (hflag == 0)?mz->am_tmpl->amt_en:mz->am_tmpl->amt_en_html;
+        tmp_tmpl = strdup(tmpl);
+        if (tmp_tmpl == NULL) {
+	    log(ERR_MEMORY_ALLOCATE, "add_message", "tmp_tmpl", strerror(errno));
+            return NULL;
+        }
+    } else if (strcasecmp(charset, "iso-2022-jp") == 0
+               || strcasecmp(charset, "\"iso-2022-jp\"") == 0) {
+        // case JIS
+        ret = dg_str2code(tmpl, &tmp_tmpl, "SJIS", "ISO-2022-JP");
+        switch (ret) {
+        case -1:
+            // memory error
+	    log(ERR_MEMORY_ALLOCATE, "add_message", "tmp_tmpl", strerror(errno));
+            return NULL;
+        case 1:
+            // failed to encode
+	    log("Failed to encode charset : add_message");
+            return NULL;
+        }
+    } else if (strcasecmp(charset, "utf-8") == 0
+               || strcasecmp(charset, "utf8") == 0
+               || strcasecmp(charset, "\"utf-8\"") == 0
+               || strcasecmp(charset, "\"utf8\"") == 0) {
+        // case UTF-8
+        ret = dg_str2code(tmpl, &tmp_tmpl, "SJIS", "UTF-8");
+        switch (ret) {
+        case -1:
+            // memory error
+	    log(ERR_MEMORY_ALLOCATE, "add_message", "tmp_tmpl", strerror(errno));
+            return NULL;
+        case 1:
+            // failed to encode
+	    log("Failed to encode charset : add_message");
+            return NULL;
+        }
+    } else {
+        // case other includes SJIS
+        tmp_tmpl = strdup(tmpl);
+        if (tmp_tmpl == NULL) {
+	    log(ERR_MEMORY_ALLOCATE, "add_message", "tmp_tmpl", strerror(errno));
+            return NULL;
+        }
+    }
+
+    // encode transfer encodings
+    switch (c_enc) {
+    case GMIME_CONTENT_ENCODING_BASE64:
+        // decode original message
+        ret = decode_b64(orig_msg, &tmp_msg);
+        if (ret < 0) {
+            // memory error
+	    log(ERR_MEMORY_ALLOCATE, "add_message", "new_tmpl", strerror(errno));
+            free(tmp_tmpl);
+            return NULL;
+        }
+        // get length
+        olen = strlen(tmp_msg);
+        tlen = strlen(tmp_tmpl);
+        nlen = olen + tlen;
+
+        // build new message
+        tmp_msg2 = (char *)malloc(nlen + 1);
+        if (tmp_msg2 == NULL) {
+            log(ERR_MEMORY_ALLOCATE, "add_message", "tmp_msg2", strerror(errno));
+            free(tmp_tmpl);
+            free(tmp_msg);
+            return NULL;
+        }
+        strcpy(tmp_msg2, tmp_tmpl);
+        strcpy((tmp_msg2 + tlen), tmp_msg);
+        free(tmp_tmpl);
+        free(tmp_msg);
+
+        new_msg = encode_b64(tmp_msg2);
+        if (new_msg == NULL) {
+            // memory error
+	    log(ERR_MEMORY_ALLOCATE, "add_message", "new_tmpl", strerror(errno));
+            free(tmp_msg2);
+            return NULL;
+        }
+        free(tmp_msg2);
+        break;
+    case GMIME_CONTENT_ENCODING_QUOTEDPRINTABLE:
+        // QP encode template 
+        new_tmpl = encode_qp(tmp_tmpl);
+        if (new_tmpl == NULL) {
+            log(ERR_MEMORY_ALLOCATE, "add_message", "new_tmpl", strerror(errno));
+            free(tmp_tmpl);
+            return NULL;
+        }
+        free(tmp_tmpl); 
+
+        // get length
+        olen = strlen(orig_msg);
+        tlen = strlen(new_tmpl);
+        nlen = olen + tlen;
+
+        // build new message
+        new_msg = (char *)malloc(nlen + 1);
+        if (new_msg == NULL) {
+            log(ERR_MEMORY_ALLOCATE, "add_message", "new_msg", strerror(errno));
+            free(tmp_tmpl);
+            free(tmp_msg);
+            return NULL;
+        }
+        strcpy(new_msg, new_tmpl);
+        strcpy((new_msg + tlen), orig_msg);
+        free(new_tmpl);
+        break;
+    default:
+        // unnecessary to encode
+
+        // get length
+        olen = strlen(orig_msg);
+        tlen = strlen(tmp_tmpl);
+        nlen = olen + tlen;
+
+        // build new message
+        new_msg = (char *)malloc(nlen + 1);
+        if (new_msg == NULL) {
+    	log(ERR_MEMORY_ALLOCATE, "add_message", "new_msg", strerror(errno));
+            free(new_tmpl);
+            return NULL;
+        }
+        strcpy(new_msg, tmp_tmpl);
+        strcpy((new_msg + tlen), orig_msg);
+        free(tmp_tmpl);
+        break;
+    }
+
+
+    return new_msg;
+}
+
+int
+replace_mime_part(GMimeObject *att, GMimeObject *object, int pos,
+                  struct mailzip *mz, GMimeContentType *ctype)
+{
+    GMimeStream *stream;
+    GMimeDataWrapper *content;
+    GMimePart *mime_part;
+    char *orig_body;
+    char *head;
+    GMimeObject *replaced;
+    const char *charset;
+    const char *c_ent;
+    GMimeContentEncoding c_enc;
+    char *new_body;
+
+    mime_part = g_mime_part_new_with_type(ctype->type, ctype->subtype);
+    charset = g_mime_object_get_content_type_parameter(att, "charset");
+    if (charset != NULL) {
+        g_mime_object_set_content_type_parameter((GMimeObject *)mime_part, "charset", charset);
+    }
+    c_ent = g_mime_object_get_header(att, "content-transfer-encoding");
+    if (c_ent != NULL) {
+        g_mime_object_append_header((GMimeObject *)mime_part, "content-transfer-encoding", c_ent);
+    }
+    c_enc = g_mime_part_get_content_encoding((GMimePart *)att);
+    orig_body = g_mime_object_to_string(att);
+    if (orig_body == NULL) {
+        // memory error
+	log(ERR_MEMORY_ALLOCATE, "replace_mime_part", "orig_body", strerror(errno));
+        return -1;
+    }
+    head = get_body_head(orig_body);
+
+    // Add message
+    new_body = add_message(head, mz, (char *)charset, c_enc, ctype);
+    if (new_body == NULL) {
+        // memory error
+	log(ERR_MEMORY_ALLOCATE, "replace_mime_part", "new_body", strerror(errno));
+        return -1;
+    }
+    free(orig_body);
+
+    stream = g_mime_stream_mem_new_with_buffer(new_body, strlen(new_body));
+    content = g_mime_data_wrapper_new_with_stream(stream, c_enc);
+    g_object_unref(stream);
+    g_mime_part_set_content_object(mime_part, content);
+    g_object_unref(content);
+    replaced = g_mime_multipart_replace((GMimeMultipart *)object, pos, (GMimeObject *)mime_part);
+    g_object_unref(mime_part);
+    g_object_unref(replaced);
+    g_mime_multipart_get_part((GMimeMultipart *)object, pos);
+    free(new_body);
+
+    return 0;
+}
+#endif	// __CUSTOMIZE2018
